/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: map.h,v 1.32 2007/07/25 02:07:49 zlb Exp $ */

#ifndef PHG_MAP_H

/* The MAP struct for handling DOFs ==> VEC mapping
 *
 * - The DOF index of an unknown is just it's natural numbering in the DOF
 *
 * - The local indices are obtained by concatenating the DOF indices, one
 *   DOF after another.
 *
 * - The vector indices are first computed (in build_L2Vmap()) as follows:
 *
 *   If nprocs == 1, the DOF indices and vector indices are identical.
 *
 *   Otherwise let verts_owned, edges_owned, faces_owned, and elems_owned
 *   be respectively the number of vertices, edges, faces, and elements
 *   owned by the submesh (with verts_owned = 0 if no vertex DOF present,
 *   etc.), then the total number of vertices, edges, faces and elements
 *   owned by the submesh is:
 *
 *	total_owned = verts_owned + edges_owned + faces_owned + elems_owned
 *
 *   and the unknowns are numbered in the order: vertices, edges, faces, and
 *   elements (i.e., all unknowns on a vertex(/edge/face/element) are numbered
 *   continuously), i.e.:
 *
 *	all DOF of map->dofs[0..ndof-1] at the first vertex owned,
 *	... ...
 *	all DOF of map->dofs[0..ndof-1] at the last vertex owned,
 *
 *	all DOF of map->dofs[0..ndof-1] at the first edge owned,
 *	... ...
 *	all DOF of map->dofs[0..ndof-1] at the last edge owned,
 *
 *	all DOF of map->dofs[0..ndof-1] at the first face owned,
 *	... ...
 *	all DOF of map->dofs[0..ndof-1] at the last face owned,
 *
 *	all DOF of map->dofs[0..ndof-1] at the first element owned,
 *	... ...
 *	all DOF of map->dofs[0..ndof-1] at the last element owned,
 *
 *   Thus if nprocs > 1 the DOFs are interleaved in the vector with the
 *   procedure above.
 *
 *   Than, as a last step, if the macro INTERLEAVE_DOFS in the file map.c
 *   is FALSE, the vector indices are reordered such that the locally
 *   owned DOF data are concatenated, i.e., dofs[0] comes the first,
 *   followed by dofs[1], etc.
 *
 *   WARNING:
 *   The mapping DOF -> VEC is different for nprocs == 1 and for nprocs > 1
 *   if INTERLEAVE_DOFS is set to TRUE.
 *
 * - The number of unknowns referenced by the submesh is denoted by localsize,
 *   and the number of unknowns owned by the submesh is denoted by nlocal.
 *   The array L2Vmap[localsize] maps a local index to its vector index,
 *   in the range [0, nlocal - 1] if the unknown is owned by the submesh,
 *   and in the range [nlocal, localsize - 1] if the unknown is not owned
 *   by the submesh (off-process entry), for the latter case, the array
 *   O2Gmap[localsize - nlocal] is used to further map it to its global index.
 *
 * The maps between different kinds of indices are illustrated by the following
 * diagram, when nprocs > 1:
 *
 *		      phgDofMapE2D		    phgMapD2L
 *	elem index -----------------> DOF index ----------------> local index
 *						(solver->offsets)
 *
 *		    L2Vmap			O2Gmap
 *	local index ------> local vector index ---------> global vector index
 *					       partition
 */
typedef struct MAP_ {
    FLOAT	lif;		/* load imbalance factor */
    DOF		**dofs;		/* list of DOFs (unknowns) */

    struct MAP_	*map_nb;	/* the map with bdry entries removed */

    INT		*L2Vmap;	/* local index to local vector index map */
    INT		*O2Gmap;	/* local-vector to global index map for
				   off-process entries */
    INT		*ordering;	/* O2Gmap[ordering[.]] is increasing */
    INT		*partition;	/* nprocs + 1 entries, vector partitioning */

    /* internal variables (partial sums of solver->dofs[0..i]->count_xxxx) */
    INT		*offset;	/* offset[i] = first local index of dofs[i] */
    INT		*count_vert, *count_edge, *count_face, *count_elem;

    INT		*bdry;		/* sorted vector indices of boundary DOF */
    INT		*bdry_partition;/* partitioning of boundary rows */
    INT		bdry_nlocal;	/* number of boundary DOF owned */
    INT		bdry_localsize;	/* total number of local boundary DOF */
    INT		bdry_nglobal;	/* global number of boundary DOF */

    INT		nlocal;		/* local matrix/vector size */
    INT		nglobal;	/* global matrix/vector size */
    INT		localsize;	/* local size, including off-proc entries */
    INT		serial_no;	/* current serial number of the grid */

    int		rank, nprocs;

    int		ref_count;	/* reference count */
    BYTE	ndof;		/* number of DOFs in 'dofs[]' */
} MAP;

#ifdef __cplusplus
extern "C" {
#endif

INT phgMapL2G(MAP *map, INT index);
#define phgMapD2L(map, dof_no, index) ((map)->offset[dof_no] + index)
#define phgMapE2L(map, dof_no, e, index) \
	phgMapD2L(map, dof_no, phgDofMapE2D((map)->dofs[dof_no], e, index))
#define phgMapE2G(map, dof_no, e, index) \
	phgMapL2G(map, phgMapE2L(map, dof_no, e, index))
#define phgMapL2V(map, index) \
	((map)->L2Vmap == NULL ? (index) : (map)->L2Vmap[index])

MAP *phgMapCreateSimpleMap(GRID *g, INT n, INT N);
MAP *phgMapCreateN(int ndof, DOF **dofs);
MAP *phgMapCreate(DOF *u, ...);
void phgMapDestroy(MAP **map_ptr);

MAP *phgMapRemoveBoundaryEntries(MAP *map);

int phgMapLocalDataToDof(MAP *map, int ndof, DOF **dofs, FLOAT *data);
int phgMapDofToLocalData(MAP *map, int ndof, DOF **dofs, FLOAT *data);

VEC *phgMapDofArraysToVecN(MAP *map, int nvec, BOOLEAN remove_bdry,
				VEC **vecptr, int ndof, DOF ***dofs);
VEC *phgMapDofArraysToVec(MAP *map, int nvec, BOOLEAN remove_bdry,
				VEC **vecptr, DOF **u, ...);
void phgMapVecToDofArraysN(MAP *map, VEC *vec, BOOLEAN remove_bdry,
				int ndof, DOF ***dofs);
void phgMapVecToDofArrays(MAP *map, VEC *vec, BOOLEAN remove_bdry, DOF **u,...);

VEC *phgMapCreateVec(MAP *map, int nvec);
MAT *phgMapCreateMat(MAP *rmap, MAP *cmap);
MAT *phgMapCreateMatrixFreeMat(MAP *rmap, MAP *cmap, MV_FUNC mv_func,
	void *mv_data, ...);

#ifdef __cplusplus
}
#endif

#ifdef NEED_GetVariableArgs
# define MAX_ARGS 255
# define GetVariableArgs(narg, args, u, type) {				\
    va_list ap;								\
									\
    (args) = phgAlloc(MAX_ARGS * sizeof(*(args)));			\
    va_start(ap, u);							\
    for ((narg) = 0; (narg) < MAX_ARGS; (narg)++) {			\
	if (u == NULL)							\
	    break;							\
	args[narg] = u;							\
	u = va_arg(ap, type);						\
    }									\
    va_end(ap);								\
    (args) = phgRealloc(args, (narg) * sizeof(*(args)));		\
}
#endif	/* NEED_GetVariableArgs */

#define PHG_MAP_H
#endif
