/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: phg.h,v 1.57 2007/05/22 07:17:31 zlb Exp $ */

#ifndef PHG_PHG_H

/*#ifdef HAVE_CONFIG_H */
#include "config.h"
/*#endif*/
#include <stdio.h>
#include "expression.h"

#ifndef __GNUC__
# define inline
#endif

#define PHG_VERSION		"0.6.0"		/**< Version string */
#define PHG_VERSION_MAJOR	0		/**< Major version number */
#define PHG_VERSION_MINOR	6		/**< Minor version number */
#define PHG_VERSION_SUBMINOR	0		/**< Subminor version number */

#ifndef Dim
# define Dim		3		/**< The Space dimensions */
#endif

#define NVert		(Dim + 1)	/**< Number of vertices / element */
#define NFace		NVert		/**< Number of faces / element */
#if Dim == 3
# define NEdge		6		/**< Number of edges / element */
#else
# define NEdge		NFace
#endif

/* elementary data types */

#if USE_LONG_DOUBLE
 typedef long double	FLOAT;		/**< real, double or float */
#else	/* USE_LONG_DOUBLE */
 typedef double		FLOAT;		/**< real, double or float */
#endif	/* USE_LONG_DOUBLE */
typedef signed int	INT;		/**< integer, at least 32 bits */
typedef unsigned int	UINT;		/**< unsigned integer */
typedef signed short	SHORT;		/**< short, at least 16 bits */
typedef unsigned short	USHORT;		/**< unsigned short */
typedef signed char	CHAR;		/**< character */
typedef unsigned char	BYTE;		/**< byte */

/* some mathematical functions, compatible with the FLOAT type */
#if USE_LONG_DOUBLE
# define Pow(a,b)	powl((FLOAT)(a),(FLOAT)(b))
# define Sqrt(a)	sqrtl((FLOAT)(a))
# define Fabs(a)	fabsl((FLOAT)(a))
# define Log(a)		logl((FLOAT)(a))
# define Exp(a)		expl((FLOAT)(a))
# define Sin(a)		sinl((FLOAT)(a))
# define Asin(a)	asinl((FLOAT)(a))
# define Cos(a)		cosl((FLOAT)(a))
# define Acos(a)	acosl((FLOAT)(a))
# define Tan(a)		tanl((FLOAT)(a))
# define Atan(a)	atanl((FLOAT)(a))
# else	/* USE_LONG_DOUBLE */
# define Pow(a,b)	pow((FLOAT)(a),(FLOAT)(b))
# define Sqrt(a)	sqrt((FLOAT)(a))
# define Fabs(a)	fabs((FLOAT)(a))
# define Log(a)		log((FLOAT)(a))
# define Exp(a)		exp((FLOAT)(a))
# define Sin(a)		sin((FLOAT)(a))
# define Asin(a)	asin((FLOAT)(a))
# define Cos(a)		cos((FLOAT)(a))
# define Acos(a)	acos((FLOAT)(a))
# define Tan(a)		tan((FLOAT)(a))
# define Atan(a)	atan((FLOAT)(a))
#endif	/* USE_LONG_DOUBLE */

enum {FALSE, TRUE};			/**< 0 and 1 */
typedef int BOOLEAN;			/**< boolean */

/* Geometric types. Also used for refinement types: EDGE means the refinement
   edge is an edge of the reference parallelepiped, FACE means the refinement
   edge is the diagonal of a face of the reference parallelepiped, and
   DIAGONAL means the refinement edge is the diagonal of the reference
   parallelepiped */
#define GTypeName(t) (					\
	((t) == VERTEX)		? "VERTEX" :		\
	((t) == EDGE)		? "EDGE" :		\
	((t) == FACE)		? "FACE" :		\
	((t) == DIAGONAL)	? "DIAGONAL" :		\
	((t) == OPPOSITE)	? "OPPOSITE" :		\
	((t) == MIXED)		? "MIXED" : "UNKNOWN")
enum {VERTEX, EDGE, FACE, DIAGONAL, OPPOSITE, MIXED};
typedef BYTE GTYPE;	/* force to BYTE to save memory */

#define ELEMENT DIAGONAL

/* types of faces (REMOTE means it has a remote neighbour), the type
   UNREFERENCED means the object does not belong to the leaf elements.

   Note:
   1. multiple bits may be set, e.g., DIRICHLET | REMOTE | INTERIOR
   2. REMOTE bit set ==> INTERIOR bit set
*/
#define BTypeName(t) (				\
	((t) & UNREFERENCED)	? "unused" :	\
	((t) & INTERIOR)	? (((t) & REMOTE) ? "remote" : "interior") : \
	((t) & DIRICHLET)	? "Dirichlet":	\
	((t) & NEUMANN)		? "Neumann" :	\
	((t) & BDRY_USER1)	? "user 1" :	\
	((t) & BDRY_USER2)	? "user 2" :	\
	((t) & OWNER)		? "owner" : "undefined")
enum {
    UNREFERENCED= 0,	/**< not referenced by leaf elements of the submesh */
    OWNER	= 1,	/**< assigned to current submesh */
    INTERIOR    = 2,	/**< interior of the mesh */
    REMOTE	= 4,	/**< shared with a remote process */
    DIRICHLET   = 8,	/**< belongs to a Dirichlet boundary */
    NEUMANN	= 16,	/**< belongs to a Neumann boundary */
    BDRY_USER1	= 32,	/**< belongs to a user type 1 boundary */
    BDRY_USER2	= 64,	/**< belongs to a user type 2 boundary */
    UNDEFINED	= 128	/**< belongs to a boundary but has no bdry type */
};
typedef BYTE BTYPE;	/**< boundary type, force to BYTE to save memory */

#define BDRY_MASK (DIRICHLET | NEUMANN | BDRY_USER1 | BDRY_USER2 | UNDEFINED)
#define IS_BDRY(b)	((b) & BDRY_MASK)

/* GRID flags */
enum {
    VERT_FLAG = 1,
    EDGE_FLAG = 2,
    FACE_FLAG = 4,
    ELEM_FLAG = 8,
    GEOM_FLAG = 16
};

typedef FLOAT COORD[Dim];		/* space coordinates */

/**
 @brief The SIMPLEX struct
 */
typedef struct SIMPLEX_ {
    struct SIMPLEX_	*children[2];		/**< Pointers to children */
    void		*neighbours[NFace];	/**< Pointers to neighbours.
						   It stores the
						   index in g->neighbours.list
						   if the neighbour is remote */
#if 0
    void		*userdata;		/**< Pointer to userdata, for
						   use by applications */
#endif
    void		*parent;		/* pointer to parent */
    INT			verts[NVert];		/**< List of vertices */
    INT			edges[NEdge];		/**< List of edges */
#if (Dim == 3)
    INT			faces[NFace];		/**< List of faces */
#endif
    INT			index;			/**< Element index */
    SHORT		mark;			/**< Refine/coarsen mark */
#if ALLOW_CURVED_BOUNDARY
    /* Note: storing the indices vertex-wise instead of edge-wise saves
       memory, but it has problems when dealing with vertices across the
       interface of two surfaces */
    CHAR		bound_func[NEdge];	/**< Indices to curved boundary
						   edges (-1 if not curved) */
#endif
    BTYPE		bound_type[NFace];	/**< Boundary types */
    BYTE		ordering;		/**< Ordering of vertices:
						   - bits 0-1: 1st vertex
						   - bits 2-3: 2nd vertex
						   - bits 4-5: 3rd vertex
						   - bits 6-7: 4th vertex */
    CHAR		generation;		/**< Generation (level) */
    BYTE		flag;			/**< Scratch flag */
#if (Dim == 3)
    GTYPE		type;			/**< Refinement type */
#endif
} SIMPLEX;

# define GlobalVertex(g,no) (no)
# define GlobalEdge(g,no) (no)
# define GlobalFace(g,no) (no)
# define GlobalElement(g,no) (no)

#define ForAllElements(g, e) \
    for (_phg_i = 0; \
	 (e = (_phg_i < (g)->nleaf) ? (g)->leaves[_phg_i] : NULL) != NULL; \
	 _phg_i++)

/**
 @brief The GRID struct
 */
#define nleaf_global	nelem_global
typedef struct {
    FLOAT	lif;		/**< Load imbalance factor */
    FLOAT	bbox[2][Dim];	/**< BoundingBox of the mesh */
    COORD	*verts;		/**< List of vertices */
    SIMPLEX	*roots;		/**< Root elements */
    SIMPLEX	**leaves;	/**< Pointers to leaf elements
				   (for mesh traversal) */
    struct DOF_	*geom;		/**< DOF containing the geometric data,
				   maintained by geom.c */
    struct DOF_	**dof;		/**< NULL terminated list of DOFs
				   defined on the mesh */
#if ALLOW_CURVED_BOUNDARY
    EXPR	**bdry_funcs;	/**< NULL terminated list of projection
				   functions for curved boundaries */
#endif
    BYTE	*types_vert;	/**< Types of vertices (bit flags) */
    BYTE	*types_edge;	/**< Types of edges (bit flags) */
    BYTE	*types_face;	/**< Types of faces (bit flags) */
    BYTE	*types_elem;	/**< Types of elements (bit flags) */
    INT		(*perlist)[2];	/**< list of pairs of periodic vertices */
    INT		nleaf;		/**< Number of leaf elements in the submesh */
    INT		nvert;		/**< Number of vertices in the submesh */
    INT		nedge;		/**< Number of edges in the submesh */
#if Dim == 3
    INT		nface;		/**< Number of faces in the submesh */
#endif
    INT		nelem;		/**< Number of element indices in the submesh */
    /* 2D, V:E:T \approx 1:3:2. 3D, V:E:F:T \approx 1:6:10:5 */
    INT		nvert_global;	/**< Number of vertices in the global mesh */
    INT		nedge_global;	/**< Number of edges in the global mesh */
#if Dim == 3
    INT		nface_global;	/**< Number of faces in the global mesh */
#endif
    INT		nelem_global;	/**< Number of tetrahedra in the global mesh */

    INT		nroot;		/**< Number of root elements */
    INT		ntree;		/**< Total number of elements in the tree */
    INT		serial_no;	/**< Number of grid changes */
    int		rank;		/**< Process rank (submesh no) */
    int		nprocs;		/**< Number of processes for the mesh */
    BOOLEAN	vert_flag;	/**< Whether use boundary type of vertices */
    BOOLEAN	edge_flag;	/**< Whether use bdry type & index of edges */
    BOOLEAN	face_flag;	/**< Whether use index of faces */
    BOOLEAN	elem_flag;	/**< Whether use index of elements */
    BOOLEAN	geom_flag;	/**< Whether use geometric data */
    BOOLEAN	periodic;	/**< Whether have periodic BC */
} GRID;

#include <float.h>
#define IsZero(a) (fabs(a) < 10.0 * DBL_MIN)
#define IsOne(a)  (fabs(a - 1.0) < 10.0 * DBL_EPSILON)

#include "utils.h"
#include "distribute.h"
#include "refine.h"
#include "coarsen.h"
#include "phg-mpi.h"
#include "albert.h"
#include "medit.h"
#include "dof.h"
#include "lagrange.h"
#include "nedelec.h"
#include "hfeb.h"
#include "hierarchical-basis.h"
#include "dg.h"
#include "quad.h"
#include "quad-gauss.h"
#include "geom.h"
#include "matvec.h"
#include "map.h"
#include "solver.h"
#include "vtk.h"
#include "opendx.h"
#include "option.h"
#include "perf.h"
#include "eigen.h"
#include "dof-utils.h"
#include "mark.h"

#define PHG_PHG_H
#endif
