/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver-trilinos.cxx,v 1.11 2007/04/13 06:47:08 zlb Exp $ */

/* Trilinos linear solvers interface */

#include "phg.h"

#if USE_TRILINOS	/* whole file */

#include <Epetra_ConfigDefs.h>
#include <Epetra_MpiComm.h>
#include <Epetra_Map.h>
#include <Epetra_Vector.h>
#include <Epetra_CrsMatrix.h>
#include <Epetra_LinearProblem.h>
#include <AztecOO.h>
#include <ml_include.h>
#include <ml_MultiLevelOperator.h>
#include <ml_epetra_utils.h>

#include <stdlib.h>

// Note: see TRILINOS_INCLUDE/az_aztec_defs.h for parameters

static const char *solver_name[] = {
    "cg", "gmres", "cgs", "tfqmr", "bicgstab",
    "slu", "symmlq", "gmresr", "fixed_pt", "analyze",
    "lu", "cg_condnum", NULL};
static int solver_list[] = {
    AZ_cg, AZ_gmres, AZ_cgs, AZ_tfqmr, AZ_bicgstab,
    AZ_slu, AZ_symmlq, AZ_GMRESR, AZ_fixed_pt,
    AZ_analyze, AZ_lu, AZ_cg_condnum};
static int solver_index = 0;

static const char *pc_name[] = {
    "none", "jacobi", "sym_gs", "neumann", "ls", "ilu", "bilu", "lu",
    "icc", "ilut", "rilu", "recursive", "smoother", "dom_decomp",
// Note: borrow the keyword AZ_multilevel for the ML preconditioner
    "ml", NULL};
static int pc_list[] = {
    AZ_none, AZ_Jacobi, AZ_sym_GS, AZ_Neumann, AZ_ls, AZ_ilu, AZ_bilu, AZ_lu,
    AZ_icc, AZ_ilut, AZ_rilu, AZ_recursive, AZ_smoother, AZ_dom_decomp,
    AZ_multilevel};
static int pc_index = 1;	/* default to Jacobi PC */

typedef struct {
    Epetra_MpiComm			*comm;
    Epetra_Map				*map;
    Epetra_Vector			*U, *B;
    Epetra_CrsMatrix			*A;
    Epetra_LinearProblem		*linsys;
    AztecOO				*solver;
    ML_Epetra::MultiLevelOperator	*mlop;
    ML_Aggregate			*agg_object;
    ML					*ml_handle;
    BOOLEAN				assembled;
} OEM_DATA;

/* convenience macro */
#define _t	((OEM_DATA *)solver->oem_data)

static int
RegisterOptions(void)
{
    phgOptionsRegisterTitle("\nTrilinos AztecOO options:", NULL);
    phgOptionsRegisterKeyword("aztec_solver", "AztecOO solver type",
				solver_name, &solver_index);
    phgOptionsRegisterKeyword("aztec_pc", "AztecOO PC type",
				pc_name, &pc_index);

    return 0;
}

#define Initialize	NULL
#define Finalize	NULL

static int
Create(SOLVER *solver)
{
    solver->oem_data = phgCalloc(1, sizeof(OEM_DATA));
    _t->assembled = FALSE;
}

static int
Destroy(SOLVER *solver)
{
    if (solver->oem_data == NULL)
	return 0;

    if (_t->mlop != NULL)
	delete _t->mlop;
    if (_t->agg_object != NULL)
	ML_Aggregate_Destroy(&_t->agg_object);
    if (_t->ml_handle != NULL)
	ML_Destroy(&_t->ml_handle);
    if (_t->solver != NULL)
	delete _t->solver;
    if (_t->linsys != NULL)
	delete _t->linsys;
    if (_t->A != NULL)
	delete _t->A;
    if (_t->B != NULL)
	delete _t->B;
    if (_t->U != NULL)
	delete _t->U;
    if (_t->map != NULL)
	delete _t->map;
    if (_t->comm != NULL)
	delete _t->comm;

    phgFree(solver->oem_data);
    solver->oem_data = NULL;

    return 0;
}

#define AddMatrixEntries	NULL
#define AddRHSEntries		NULL
 
static int
Assemble(SOLVER *solver)
{
    MAP *map0 = solver->rhs->map;
    int Nglobal = map0->nglobal, Nlocal = map0->nlocal;
    int i, n0, nmax;
    MAT_ROW *row;
#ifndef DONT_HAVE_G_COMM
    int j, k, *cols;
#endif  /* DONT_HAVE_G_COMM */

    if (_t->assembled == TRUE)
	return 0;

    _t->assembled = TRUE;
    assert(solver->mat->assembled);
    assert(solver->mat != NULL);
    if (solver->mat->type == PHG_PACKED)
	phgMatUnpack(solver->mat);
    assert(solver->mat->type == PHG_UNPACKED);
    assert(solver->sent_to_oem == FALSE);

#ifdef DONT_HAVE_G_COMM
    _t->comm = new Epetra_MpiComm(MPI_COMM_SELF);
#else	/* DONT_HAVE_G_COMM */
    _t->comm = new Epetra_MpiComm(map0->comm);
#endif	/* DONT_HAVE_G_COMM */

    n0 = map0->partition[map0->rank];

    /* create map */
    int *work = new int[Nlocal];
    for (i = 0; i < Nlocal; i++)
	work[i] = i + n0;
    _t->map = new Epetra_Map(Nglobal, Nlocal, work, 0, *_t->comm);

    /* create vectors */
    _t->U = new Epetra_Vector(*_t->map);
    _t->B = new Epetra_Vector(*_t->map);

    /* create matrix */
    nmax = 0;
    for (i = 0, row = solver->mat->rows; i < Nlocal; i++, row++) {
	work[i] = row->ncols;
	if (nmax < row->ncols)
	    nmax = row->ncols;
    }
    _t->A = new Epetra_CrsMatrix(Copy, *_t->map, work);
    delete [] work;
#ifndef DONT_HAVE_G_COMM
    cols = (int *)phgAlloc(nmax * sizeof(*cols));
#endif  /* DONT_HAVE_G_COMM */
    /* send matrix entries */
    assert(sizeof(int) == sizeof(INT) && sizeof(double) == sizeof(FLOAT));
    for (i = 0, row = solver->mat->rows; i < Nlocal; i++, row++) {
	if (row->ncols == 0)
	    continue;
#ifdef DONT_HAVE_G_COMM
	_t->A->InsertGlobalValues(n0 + i, row->ncols, row->data, row->cols);
#else	/* DONT_HAVE_G_COMM */
	for (j = 0; j < row->ncols; j++) {
	    k = row->cols[j];
	    cols[j] = (k < Nlocal ? k + n0 : solver->mat->O2Gmap[k - Nlocal]);
	}
	_t->A->InsertGlobalValues(n0 + i, row->ncols, row->data, cols);
#endif  /* DONT_HAVE_G_COMM */
	if (solver->mat->ref_count == 0) {
	    phgFree(row->data);
	    row->data = NULL;
	    phgFree(row->cols);
	    row->cols = NULL;
	    row->ncols = 0;
	}
    }
    if (solver->mat->ref_count == 0)
	phgMatFreeMatrix(solver->mat);
    solver->sent_to_oem = TRUE;
#ifndef DONT_HAVE_G_COMM
    phgFree(cols);
#endif  /* DONT_HAVE_G_COMM */
    _t->A->FillComplete();

    /* define the linear problem */
    _t->linsys = new Epetra_LinearProblem(_t->A, _t->U, _t->B);
    _t->solver = new AztecOO(*_t->linsys);

    // set solver options
    int *options = new int[AZ_OPTIONS_SIZE];
    double *params = new double[AZ_PARAMS_SIZE];
    // AZ_defaults(options, params);

    /* retrieve internal options and parameters */
    memcpy(options, _t->solver->GetAllAztecOptions(),
			AZ_OPTIONS_SIZE * sizeof(*options));
    memcpy(params, _t->solver->GetAllAztecParams(),
			AZ_PARAMS_SIZE * sizeof(*params));

    if (phgVerbosity <= 0)
	options[AZ_output]	= AZ_none;
    options[AZ_conv]		= AZ_Anorm;
    options[AZ_solver]		= solver_list[solver_index];

    int levels = 10;	/* ML levels */
    switch (pc_list[pc_index]) {
	case AZ_multilevel:	/* set up ML preconditioner */
	    ML_Set_PrintLevel(phgVerbosity);
	    ML_Create(&_t->ml_handle, levels);
	    EpetraMatrix2MLMatrix(_t->ml_handle, 0, _t->A);
	    ML_Aggregate_Create(&_t->agg_object);
	    levels = ML_Gen_MGHierarchy_UsingAggregation(_t->ml_handle,
			0/*finest level*/, ML_INCREASING, _t->agg_object);
	    ML_Gen_Smoother_SymGaussSeidel(_t->ml_handle, ML_ALL_LEVELS,
			ML_BOTH, 1, ML_DEFAULT);
	    ML_Gen_Solver(_t->ml_handle, ML_MGV, 0, levels - 1);
	    _t->mlop = new ML_Epetra::MultiLevelOperator(_t->ml_handle,
			*_t->comm, *_t->map, *_t->map);
	    _t->solver->SetPrecOperator(_t->mlop);
	    break;
	case AZ_ilu:
	    options[AZ_precond] = AZ_dom_decomp;
	    options[AZ_subdomain_solve] = AZ_ilu;
	case AZ_lu:
	    options[AZ_precond] = AZ_dom_decomp;
	    options[AZ_subdomain_solve] = AZ_lu;
	case AZ_ilut:
	    options[AZ_precond] = AZ_dom_decomp;
	    options[AZ_subdomain_solve] = AZ_ilut;
	    break;
	default:
	    options[AZ_precond] = pc_list[pc_index];
    }
    _t->solver->SetAllAztecOptions(options);
    _t->solver->SetAllAztecParams(params);

    delete [] params;
    delete [] options;

    return 0;
}

static int
Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    int i;
    double *values;

    Assemble(solver);

    /* set RHS */
    _t->B->ExtractView(&values);
    for (i = 0; i < solver->rhs->map->nlocal; i++)
	values[i] = (double)solver->rhs->data[i];

    /* set initial solution */
    _t->U->ExtractView(&values);
    for (i = 0; i < solver->rhs->map->nlocal; i++)
	values[i] = (double)x->data[i];

    /* solve */
    _t->solver->/*recursive*/Iterate(solver->maxits, solver->rtol);

    /* get nbr iters and final residual */
    //solver->residual = _t->solver->RecursiveResidual();
    //solver->residual = _t->solver->ScaledResidual();
    solver->residual = _t->solver->TrueResidual();
    solver->nits = _t->solver->NumIters();

    /* copy solution */
    for (i = 0; i < solver->rhs->map->nlocal; i++)
	x->data[i] = values[i];

    if (destroy)
	Destroy(solver);

    return solver->nits;
}

/*--------------------------------------------------------------------*/

OEM_SOLVER phgSolverTrilinos_ = {
    "Trilinos", RegisterOptions,
    Initialize, Finalize, Create, Destroy, AddMatrixEntries, AddRHSEntries,
    Assemble, Solve
};

#endif		/* USE_TRILINOS */
