/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver.c,v 1.261 2007/06/10 01:41:47 zlb Exp $ */

#define NEED_GetVariableArgs

#include "phg.h"
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <math.h>

/* the workaround when the local size of the matrix is 0 on some processes.
 * Typical test case:
 *      mpirun -np 4 poisson -fem_type=P1 -mesh_file=tetra.dat -pre_refines=2
 */

/* list of solvers */
#define NSolvers	(sizeof(phgSolverList_)/sizeof(phgSolverList_[0]) - 1)

/* Solver names */
const char *phgSolverNames_[] = {
    "none",		/* dummy entry used by other interfaces (BLOPEX) */
    "pcg",
    "gmres",
#if USE_HYPRE
    "hypre",
#endif
#if USE_PETSC
    "petsc",
#endif
#if USE_TRILINOS
    "trilinos",
#endif
#if USE_MUMPS
    "mumps",
#endif
#if USE_SUPERLU
    "superlu",
#endif
#if USE_SPOOLES
    "spooles",
#endif
#if USE_SPC
    "spc",
#endif
#if USE_LASPACK
    "laspack",
#endif
    "builtin",
    NULL
};

/* Solvers list */
OEM_SOLVER *phgSolverList_[] = {
    NULL,		/* dummy entry reserved for other interfaces */
    SOLVER_PCG,
    SOLVER_GMRES,
#if USE_HYPRE
    SOLVER_HYPRE,
#endif
#if USE_PETSC
    SOLVER_PETSC,
#endif
#if USE_TRILINOS
    SOLVER_TRILINOS,
#endif
#if USE_MUMPS
    SOLVER_MUMPS,
#endif
#if USE_SUPERLU
    SOLVER_SUPERLU,
#endif
#if USE_SPOOLES
    SOLVER_SPOOLES,
#endif
#if USE_SPC
    SOLVER_SPC,
#endif
#if USE_LASPACK
    SOLVER_LASPACK,
#endif
    SOLVER_BUILTIN,
};
static int default_solver = -1;

/* Parameters */
static INT maxit = 10000;
static FLOAT rtol = 1e-10;
static FLOAT abstol = 1e-30;

int
phgSolverRegisterOptions(void)
{
    int i;

    phgOptionsRegisterTitle("\nLinear solver options:", NULL);
    phgOptionsRegisterKeyword("default_solver", "Default solver",
			      phgSolverNames, &default_solver);
    phgOptionsRegisterInt("solver_maxit",
			  "Default maxi. number of iterations", &maxit);
    phgOptionsRegisterFloat("solver_rtol",
			    "Default relative convergence tolerance", &rtol);
    phgOptionsRegisterFloat("solver_atol",
			    "Default absolute convergence tolerance", &abstol);

    /* Register eigensolver options */
    phgEigenSolve(NULL, NULL, 0, 0, 0.0, NULL, NULL, NULL);

    for (i = 0; i < NSolvers; i++)
	if (phgSolverList[i] != NULL &&
	    phgSolverList[i]->RegisterOptions != NULL)
	    phgSolverList[i]->RegisterOptions();

    return 0;
}

static int
send_matrix_to_oem_solver(SOLVER *solver)
/* sends data to OEM solver */
{
    MAT *mat = solver->mat;
    INT i, j, n0, max_cols;
    MAT_ROW *row;
    FLOAT total_cols;

    FunctionEntry;

    if (solver->sent_to_oem)
	Return 0;

    if (!solver->oem_created) {
	solver->oem_created = TRUE;
	solver->oem_solver->Create(solver);
    }

    /* check whether the solver has AddMatrixEntries function */
    if (mat->type == PHG_MATRIX_FREE ||
	solver->oem_solver->AddMatrixEntries == NULL)
	Return 0;

    assert(mat->type != PHG_DESTROYED);

    if (mat->type == PHG_PACKED)
	phgMatUnpack(mat);

    phgInfo(1, "sending matrix to OEM solver...\n");

    /* send matrix to OEM solver and free PHG's matrix */
    j = n0 = mat->rmap->partition[mat->rmap->rank];
    row = mat->rows;
    max_cols = total_cols = 0.0;
    for (i = 0; i < mat->rmap->nlocal; i++, row++, j++) {
	if (row->ncols <= 0)
	    phgError(1, "%s: matrix row %d is empty!\n", __func__, i);
	total_cols += row->ncols;
	if (max_cols < row->ncols) {
	    max_cols = row->ncols;
	}
	solver->oem_solver->AddMatrixEntries(solver, 1, &j, row->ncols,
					     row->cols, row->data);
	if (mat->ref_count == 0) {
	    /* free PHG matrix row */
	    phgFree(row->cols);
	    phgFree(row->data);
	    row->cols = NULL;
	    row->data = NULL;
	    row->ncols = row->alloc = 0;
	}
    }

    phgInfo(2, "max cols = %d, average cols = %lg\n", max_cols,
	    ((double)total_cols) / solver->mat->rmap->nlocal);

    if (mat->ref_count == 0)
	phgMatFreeMatrix(mat);
    solver->sent_to_oem = TRUE;

    Return solver->oem_solver->Assemble(solver);
    return 0;		/* to avoid possible gcc warnings */
}

BOOLEAN
phgSolverDenseLU(int n, FLOAT *a0, int pvt[])
/* LU factorization (partial pivoting). TODO: call LAPACK for large n */
{
    int i, j, k;
    FLOAT d, r;
#define a(i,j)	(a0[(i) * n + (j)])

    for (i = 0; i < n; i++) {
	d = Fabs(a(i,i));
	k = i;
	for (j = i + 1; j < n; j++) {
	    if ((r = Fabs(a(j,i))) > d) {
		d = r;
		k = j;
	    }
	}
	if (d == 0.0) {
#if 0 * DEBUG
	    printf("Dense solver failed: i = %d\n", i);
	    for (j = 0; j < n; j++) {
		for (k = 0; k < n; k++)
		    printf(" %9.2le", (k >= i || k >= j) ? a(j,k) : 0.);
	    }
#endif
	    return FALSE;
	}
	pvt[i] = k;
	if (k != i) {
	    /* exchange row i and row k */
	    for (j = i; j < n; j++) {
		d = a(i,j);
		a(i,j) = a(k,j);
		a(k,j) = d;
	    }
	}
	if ((d = a(i,i)) != 1.0) {
	    a(i,i) = d = 1.0 / d;
	    for (j = i + 1; j < n; j++)
		a(i,j) *= d;
	}
	for (j = i + 1; j < n; j++) {
	    if ((d = a(j,i)) == 0.0)
		continue;
	    for (k = i + 1; k < n; k++)
		a(j,k) -= d * a(i,k);
	}
    }
#undef a

    return TRUE;
}

void
phgSolverDenseSV(int n, FLOAT *a0, int pvt[], int m, FLOAT *b0)
/* solves L*U*X = B. TODO: call LAPACK for large n */
{
    int i, j, k;
    FLOAT d;
#define a(i,j)	(a0[(i) * n + (j)])
#define b(i,j)	(b0[(i) * m + (j)])

    for (i = 0; i < n; i++) {
	k = pvt[i];
	if (k != i) {
	    /* exchange row i with row k */
	    for (j = 0; j < m; j++) {
		d = b(i,j);
		b(i,j) = b(k,j);
		b(k,j) = d;
	    }
	}
	if ((d = a(i,i)) != 1.0) {
	    for (j = 0; j < m; j++)
		b(i,j) *= d;
	}
	for (j = i + 1; j < n; j++) {
	    if ((d = a(j,i)) == 0.0)
		continue;
	    for (k = 0; k < m; k++)
		b(j,k) -= d * b(i,k);
	}
    }

    for (i = n - 2; i >= 0; i--)
	for (j = i + 1; j < n; j++) {
	    d = a(i,j);
	    for (k = 0; k < m; k++)
		b(i,k) -= d * b(j,k);
	}
#undef a
#undef b

    return;
}

BOOLEAN
phgSolverDenseSolver(int n, int m, FLOAT *a0, FLOAT *b0)
/* solves AX = B. TODO: call LAPACK for large n */
{
#if 1
    int i, j, k;
    FLOAT d, r;
#define a(i,j)	(a0[(i) * n + (j)])
#define b(i,j)	(b0[(i) * m + (j)])

    for (i = 0; i < n; i++) {
	d = Fabs(a(i,i));
	k = i;
	for (j = i + 1; j < n; j++) {
	    if ((r = Fabs(a(j,i))) > d) {
		d = r;
		k = j;
	    }
	}
	if (d == 0.0) {
#if 0 * DEBUG
	    printf("Dense solver failed: i = %d\n", i);
	    for (j = 0; j < n; j++) {
		for (k = 0; k < n; k++)
		    printf(" %9.2le", (k >= i || k >= j) ? a(j,k) : 0.);
		for (k = 0; k < m; k++)
		    printf(" = %9.2le\n", b(j,k));
	    }
#endif
	    return FALSE;
	}
	if (k != i) {
	    /* exchange row i with row k */
	    for (j = i; j < n; j++) {
		d = a(i,j);
		a(i,j) = a(k,j);
		a(k,j) = d;
	    }
	    for (j = 0; j < m; j++) {
		d = b(i,j);
		b(i,j) = b(k,j);
		b(k,j) = d;
	    }
	}
	if ((d = a(i,i)) != 1.0) {
	    d = 1.0 / d;
	    for (j = i + 1; j < n; j++)
		a(i,j) *= d;
	    for (j = 0; j < m; j++)
		b(i,j) *= d;
	}
	for (j = i + 1; j < n; j++) {
	    if ((d = a(j,i)) == 0.0)
		continue;
	    for (k = i + 1; k < n; k++)
		a(j,k) -= d * a(i,k);
	    for (k = 0; k < m; k++)
		b(j,k) -= d * b(i,k);
	}
    }

    for (i = n - 2; i >= 0; i--)
	for (j = i + 1; j < n; j++) {
	    d = a(i,j);
	    for (k = 0; k < m; k++)
		b(i,k) -= d * b(j,k);
	}
#undef a
#undef b
#else
    int *pvt = phgAlloc(n * sizeof(*pvt));

    if (phgSolverDenseLU(n, a0, pvt) == FALSE)
	return FALSE;
    phgSolverDenseSV(n, a0, pvt, m, b0);
    phgFree(pvt);
#endif

    return TRUE;
}

void
phgSolverUpdateRHS(SOLVER *src, SOLVER *dst, BOOLEAN force)
/* This function computes:
 *	dst->rhs -= B(dst->mat) * D(src->mat)^(-1) * src->rhs
 * where:
 *	A(mat) = mat with all bdry columns set to 0,
 *	B(mat) = the matrix composed of the bdry columns of mat,
 *	D(mat) = diag matrix with non-bdry part = I and bdry part = diag(mat)
 *
 * Note: if mat->handle_bdry_eqns == TRUE, then after MAT assembly, only the
 * diagonal entries in the bdry rows can be nonzero, and A is stored in
 * mat->rows, B+D is stored in mat->bdry_rows (note both mat->rows and
 * mat->bdry_rows contain D). Since mat->rows might have been destroyed when
 * sending data to OEM solvers, we get D from mat->bdry_rows.
 *
 * TODO: replace bdry_rows with a MAT object, and use mat-vec operations. */
{
    MAT *mat;
    VEC *rhs;
    MAT_ROW *row;
    INT i, j, k;

    assert(src != NULL);
    if (dst == NULL)
	dst = src;

    if (dst->rhs_updated && !force)
	return;
    dst->rhs_updated = TRUE;

    if (src->mat == NULL || !src->mat->handle_bdry_eqns)
	return;

    assert(dst->mat != NULL && dst->mat->handle_bdry_eqns);
    assert(src->mat->rmap->nlocal == dst->mat->rmap->nlocal);

    if (!src->sent_to_oem && !src->mat->assembled)
	phgMatAssemble(src->mat);

    if (!dst->sent_to_oem && !dst->mat->assembled)
	phgMatAssemble(dst->mat);

    if (!src->rhs->assembled)
	phgVecAssemble(src->rhs);

    if (!dst->rhs->assembled)
	phgVecAssemble(dst->rhs);

    mat = dst->mat;
    rhs = dst->rhs;

    for (i = 0, row = mat->bdry_rows; i < mat->cmap->nlocal; i++, row++) {
	for (j = 0; j < row->ncols; j++) {
	    if ((k = row->cols[j]) == i)
		continue;		/* skip column on diagonal */
	    assert(src->mat->bdry_rows[k].ncols == 1 &&
		   src->mat->bdry_rows[k].cols[0] == k);
	    rhs->data[i] -= src->rhs->data[k] * row->data[j] /
					src->mat->bdry_rows[k].data[0];
	}
    }
}

int
phgSolverResetRHS(SOLVER *solver)
{
    assert(solver->rhs != NULL || solver->mat != NULL);
    if (solver->rhs == NULL)
	solver->rhs = phgMapCreateVec(solver->mat->rmap, 1);
    else
	bzero(solver->rhs->data, solver->rhs->nvec *
			solver->rhs->map->nlocal * sizeof(*solver->rhs->data));
    solver->rhs->assembled = FALSE;
    solver->rhs_updated = FALSE;
    return 0;
}

/*-------------------------- User interface --------------------------------*/

int
phgSolverInitialize(int *argc, char ***argv)
{
    int i = 0;

    for (i = 0; i < NSolvers; i++)
	if (phgSolverList[i] != NULL && phgSolverList[i]->Initialize != NULL)
	    phgSolverList[i]->Initialize(argc, argv);

    return 0;
}

int
phgSolverFinalize(void)
{
    int i = 0;

    for (i = 0; i < NSolvers; i++)
	if (phgSolverList[i] != NULL && phgSolverList[i]->Finalize != NULL)
	    phgSolverList[i]->Finalize();

    return 0;
}

MAT *
phgSolverGetMat(SOLVER *solver)
{
    solver->mat->ref_count++;
    return solver->mat;
}

SOLVER *
phgSolverMat2Solver(OEM_SOLVER *oem_solver, MAT *mat)
{
    SOLVER *solver = phgCalloc(1, sizeof(*solver));
    int j;

    solver->oem_solver = oem_solver;
    if (solver->oem_solver == NULL && default_solver >= 0) {
	static BOOLEAN warned = FALSE;
	solver->oem_solver = phgSolverList[default_solver];
	if (!warned && mat->rmap->rank == 0 && solver->oem_solver == NULL) {
	    warned = TRUE;
	    phgWarning("solver \"%s\" not available, will use the first "
			"available solver.\n", phgSolverNames[default_solver]);
	}
    }

    /* fall back to first available solver if the requested solver is not
     * available or is SOLVER_DEFAULT */
    if (solver->oem_solver == NULL) {
	for (j = 0; j < NSolvers; j++)
	    if (phgSolverList[j] != NULL) {
		solver->oem_solver = phgSolverList[j];
		break;
	    }
    }

    if (solver->oem_solver == NULL) {
	phgInfo(-1, "phgSolverCreate: no solver available\n");
	phgError(1, "phgSolverCreate: abort.\n");
    }

    phgInfo(1, "using linear solver \"%s\".\n", solver->oem_solver->name);

    mat->ref_count++;
    solver->mat = mat;
    solver->rhs = phgMapCreateVec(mat->rmap, 1);
    solver->rhs->mat = mat;	/* reverse pointer to the matrix */

    solver->rhs_updated = FALSE;
    solver->sent_to_oem = FALSE;
    solver->oem_created = FALSE;
    solver->oem_data = NULL;

    solver->rtol = rtol;
    solver->atol = abstol;
    solver->maxits = maxit;
    solver->pc = NULL;
    solver->pc_destroy = NULL;
    solver->warn_maxits = TRUE;

    return solver;
}

SOLVER *
phgSolverCreate(OEM_SOLVER *oem_solver, DOF *u, ...)
{
    MAP *map;
    MAT *mat;
    int ndof;
    DOF **dofs;

    GetVariableArgs(ndof, dofs, u, DOF *);

    map = phgMapCreateN(ndof, dofs);
    phgFree(dofs);
    mat = phgMapCreateMat(map, map);
    phgMapDestroy(&map);

    mat->handle_bdry_eqns = TRUE;

    mat->ref_count--;
    return phgSolverMat2Solver(oem_solver, mat);
}

int
phgSolverDestroy(SOLVER **solver_ptr)
{
    SOLVER *solver = *solver_ptr;

    if (solver == NULL)
	return 0;

    solver->oem_solver->Destroy(solver);

    if (solver->pc_destroy != NULL)
	solver->pc_destroy(solver);
    solver->pc_destroy = NULL;

    phgMatDestroy(&solver->mat);

    phgVecDestroy(&solver->rhs);

    phgFree(solver);

    *solver_ptr = NULL;
    return 0;
}

int
phgSolverVecSolve(SOLVER *solver, BOOLEAN destroy, VEC *x)
/* returns number of iterations, < 0 ==> error */
{
    INT i, j;
    FLOAT *rhs;
    size_t mem_peak = 0;

    assert(solver->mat->cmap->nlocal == x->map->nlocal);

    if (!solver->mat->assembled)
	phgMatAssemble(solver->mat);

    if (!solver->sent_to_oem)
	send_matrix_to_oem_solver(solver);

    if (!solver->rhs->assembled)
	phgVecAssemble(solver->rhs);

    if (!solver->rhs_updated)
	phgSolverUpdateRHS(solver, solver, FALSE);

    /* send RHS to OEM solver and free PHG's RHS */
    if (solver->oem_solver->AddRHSEntries != NULL) {
	j = solver->mat->rmap->partition[solver->mat->rmap->rank];
	rhs = solver->rhs->data;
	for (i = 0; i < solver->mat->rmap->nlocal; i++, j++)
	    solver->oem_solver->AddRHSEntries(solver, 1, &j, rhs++);
    }

    if (phgVerbosity >= 1) {
	mem_peak = phgMemoryPeakReset(0);
	phgInfo(0, "calling OEM solver...\n");
    }
    i = solver->oem_solver->Solve(solver, x, destroy);
    if (phgVerbosity >= 1) {
	mem_peak = phgMemoryPeakRestore(mem_peak);
	phgInfo(0, "%0.4lgMB memory used in OEM solver.\n",
		mem_peak / (1024.0 * 1024.0));
    }

    if (x->map->rank == 0) {
	if (i < 0)
	    phgError(1, "error occured in OEM solver \"%s\".\n",
		     solver->oem_solver->name);
	if (solver->warn_maxits&&i >= solver->maxits && solver->rtol > 0.0)
	    phgWarning("maxits attained in OEM solver, "
		       "the result may be inaccurate.\n");
    }

    return i;
}

int
phgSolverSolve(SOLVER *solver, BOOLEAN destroy, DOF *u, ...)
/* returns number of iterations, < 0 ==> error */
{
    int i, ndof;
    DOF **dofs;
    VEC *x;

    GetVariableArgs(ndof, dofs, u, DOF *);

    /* make sure the DOFs match those used in creating the solver */
    for (i = 0; i < ndof; i++) {
	if (dofs[i]->type != solver->mat->rmap->dofs[i]->type)
	    phgError(1,
		     "phgSolveSolve: DOF %d (%s), types mismatch: %s<=>%s\n",
		     i, dofs[i]->name, dofs[i]->type,
		     solver->mat->rmap->dofs[i]->type);
	if (dofs[i]->dim != solver->mat->rmap->dofs[i]->dim)
	    phgError(1,
		     "phgSolveSolve: DOF %d (%s), dims mismatch: %d<=>%d\n",
		     i, dofs[i]->name, dofs[i]->dim,
		     solver->mat->rmap->dofs[i]->dim);
    }

    x = phgMapCreateVec(solver->mat->cmap, 1);
    /* copy initial solution to x */
    phgMapDofToLocalData(x->map, ndof, dofs, x->data);
    /* call solver */
    i = phgSolverVecSolve(solver, destroy, x);
    /* copy solution from x */
    phgMapLocalDataToDof(x->map, ndof, dofs, x->data);
    phgFree(dofs);
    phgVecDestroy(&x);

    return i;
}

void
phgSolverSetTol(SOLVER *solver, FLOAT user_tol)
{
    solver->rtol = user_tol;
}

void
phgSolverSetMaxits(SOLVER *solver, INT user_maxits)
{
    solver->maxits = user_maxits;
}
