/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* VTK output functions
 *
 * $Id: vtk.c,v 1.13 2007/01/06 15:58:49 zlb Exp $
 */

#include "phg.h"
#include "io.h"
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>	/* PATH_MAX */
#include <ctype.h>

/* buffer for broadcasting filename */
static const char *filename;
/* buffer for broadcasting list of DOF ids */
static BYTE dof_ids[258];	/* last two bytes for filename length */

static unsigned char *buffer = NULL;
static size_t buffer_pos, buffer_size;

#define VTK_FLOAT float

static void *
swap_endian(void *buff, size_t n)
{
#if PHG_BIGENDIAN
    return buff;
#else	/* PHG_BIGENDIAN */
    static unsigned char buff0[8];
    unsigned char *p0, *p1;

    if (n <= 1) return buff;

    assert(sizeof(buff0) >= n);

    p0 = buff0 + n;
    p1 = buff;
    while (p0 > buff0) *(--p0) = *(p1++);
    return buff0;
#endif	/* PHG_BIGENDIAN */
}

static void
bigendian_copy(void *buf0, size_t size, size_t count, size_t loc)
{
    size_t i;
    char *buf = buf0;

    assert(buffer != NULL && loc + size * count <= buffer_size);
    for (i = 0; i < count; i++) {
	memcpy(buffer + loc, swap_endian(buf, size), size);
	buf += size;
	loc += size;
    }
}

#define BigEndianAppend(buf, size, count) {			\
    size_t size0 = size, count0 = count;			\
    bigendian_copy(buf, size0, count0, buffer_pos);		\
    buffer_pos += size0 * count0;				\
}

static char *
parse_name(GRID *g, int no)
{
    static char name0[128 + 16];
    char name[128];
    char *p = name;
    const char *s;
    int i, n;

    s = phgDofIndex2Pointer(g, dof_ids[no + 1])->name;

    /* check for duplicate names */
    for (i = 0, n = 0; i < no; i++)
	if (!strcmp(s, phgDofIndex2Pointer(g, dof_ids[i + 1])->name))
	    n++;

    while (*s != '\0' && p - name < sizeof(name) - 1) {
	*p = isspace((int)*s) ? '_' : *s;
	p++;
	s++;
    }
    *p = '\0';

    if (n == 0)
	sprintf(name0, "%s", name);
    else
	sprintf(name0, "%s_%d", name, n);

    return name0;
}

void
phgExportVTK_(GRID *g)
/* Output grid in VTK unstructured grid format. */
{
    int k, a, dim;
    size_t size;
    INT i;
    INT nleaf, nvert;
    SIMPLEX *s;
    DOF *dof;
    BOOLEAN has_cell_data = FALSE, has_point_data = FALSE;
    FLOAT *values;

    ParallelFunction(phgExportVTK_, TRUE);

    if (g == NULL || g->nprocs <= phgRank) return;

    nleaf = g->nleaf;
    nvert = g->nvert;

    assert(g->vert_flag);
    assert(g->rank != 0 || sizeof(a) == 4);	/* to be fixed later */

    filename = phgIOAddExtension(filename, "vtk");
    if (!phgIOOpen(g, (char *)filename)) {
	phgError(0, "cannot create VTK file \"%s\".\n", filename);
	return;
    }
    if (g->rank == 0) phgInfo(1, "creating VTK file \"%s\".\n", filename);

    buffer_size = nvert * sizeof(VTK_FLOAT) * 3;
    if ((size = nleaf * 5 * sizeof(a)) > buffer_size)
	buffer_size = size;
    /* compute size of DOF data */
    for (a = 0; a < dof_ids[0]; a++) {
	dof = phgDofIndex2Pointer(g, dof_ids[a + 1]);
        dim = DofDim(dof);
	dim = dim > 4 ? 4 : dim;
	if (dof->type == DOF_P0 || dof->type == DOF_DG0) {
	    size = nleaf * dim * sizeof(VTK_FLOAT);
	    has_cell_data = TRUE;
	}
	else {
	    size = nvert * dim * sizeof(VTK_FLOAT);
	    has_point_data = TRUE;
	}
	if (buffer_size < size)
	    buffer_size = size;
    }
    buffer = phgAlloc(buffer_size);

    /* File header and vertices */
    phgIOPrintf(g, "# vtk DataFile Version 2.0\n"
		"Tetrahedral grid created by PHG %s.\n"
		"BINARY\nDATASET UNSTRUCTURED_GRID\n"
		"POINTS %d %s\n", PHG_VERSION, g->nvert_global,
		sizeof(VTK_FLOAT) == 8 ? "double" : "float");
    buffer_pos = 0;
    for (i = 0; i < nvert; i++) {
	for (a = 0; a < 3; a++) {
	    VTK_FLOAT x = g->verts[i][a];
	    BigEndianAppend(&x, sizeof(x), 1);
	}
    }
    phgIOWrite(g, buffer, sizeof(VTK_FLOAT)*3, nvert, g->nvert_global,
		NULL, NULL);
    
    /* tetrahedra */
    phgIOPrintf(g, "\nCELLS %d %d\n", g->nelem_global, g->nelem_global * 5);
    buffer_pos = 0;
    /* element vertices */
    ForAllElements(g, s) {
	a = 4;
	BigEndianAppend(&a, sizeof(a), 1);
	for (i = 0; i < 4; i++) {
	    a = GlobalVertex(g, s->verts[i]);
	    BigEndianAppend(&a, sizeof(a), 1);
	}
    }
    phgIOWrite(g, buffer, sizeof(a) * 5, nleaf, g->nelem_global, NULL, NULL); 

    phgIOPrintf(g, "\nCELL_TYPES %d\n", g->nelem_global);
    buffer_pos = 0;
    a = 10;
    for (i = 0; i < nleaf; i++)
	BigEndianAppend(&a, sizeof(a), 1);
    phgIOWrite(g, buffer, sizeof(a), nleaf, g->nelem_global, NULL, NULL); 

    if (has_point_data) {	/* point data */
	phgIOPrintf(g, "\nPOINT_DATA %d\n", g->nvert_global);
	for (a = 0; a < dof_ids[0]; a++) {
	    DOF *dof0;
	    dof = dof0 = phgDofIndex2Pointer(g, dof_ids[a + 1]);
	    if (dof->type == DOF_P0 || dof->type == DOF_DG0)
		continue;
	    if (dof->type != DOF_P1)
		dof = phgDofCopy(dof0, NULL, DOF_P1, dof0->name);
	    if (g->rank == 0)
		phgInfo(2, "writing DOF %d: \"%s\"\n", a, dof->name);
	    dim = dof->dim;
	    if (dim > 4) {
		if (g->rank == 0)
		    phgWarning("only saving the first 4 components of \"%s\"\n",
				dof->name);
		dim = 4;
	    }
	    if (FALSE && dim == Dim)
	        phgIOPrintf(g, "\nVECTORS %s %s\nLOOKUP_TABLE default\n",
			    parse_name(g, a),
			    sizeof(VTK_FLOAT) == 8 ? "double" : "float");
	    else
	        phgIOPrintf(g, "\nSCALARS %s %s %d\nLOOKUP_TABLE default\n",
			    parse_name(g, a),
			    sizeof(VTK_FLOAT) == 8 ? "double" : "float", dim);
	    /* collect data */
#if 0
	    ForAllElements(g, s) {
		int j;
		for (k = 0; k < NVert; k++) {
		    values = DofVertexData(dof, s->verts[k]);
		    buffer_pos = s->verts[k] * dim * sizeof(VTK_FLOAT);
		    for (j = 0; j < dim; j++) {
			VTK_FLOAT v = *(values++);
			BigEndianAppend(&v, sizeof(v), 1);
		    }
		}
	    }
#else
	    buffer_pos = 0;
	    values = DofVertexData(dof, 0);
	    for (i = 0; i < dim * nvert; i++) {
		VTK_FLOAT v = *(values++);
		BigEndianAppend(&v, sizeof(v), 1);
	    }
#endif
	    phgIOWrite(g, buffer, sizeof(VTK_FLOAT) * dim, nvert,
			g->nvert_global, NULL, NULL); 
	    if (dof != dof0)
		phgDofFree(&dof);
	}
    }

    phgIOPrintf(g, "\nCELL_DATA %d\n", g->nelem_global);
    if (has_cell_data) {
	for (a = 0; a < dof_ids[0]; a++) {
	    dof = phgDofIndex2Pointer(g, dof_ids[a + 1]);
	    if (dof->type != DOF_P0 && dof->type != DOF_DG0)
		continue;
	    if (g->rank == 0)
		phgInfo(2, "writing DOF %d: \"%s\"\n", a, dof->name);
	    dim = dof->dim;
	    if (dim > 4) {
		if (g->rank == 0)
		    phgWarning("only saving the first 4 components of \"%s\"\n",
				dof->name);
		dim = 4;
	    }
	    if (FALSE && dim == Dim)
	 	phgIOPrintf(g, "\nVECTORS %s %s\nLOOKUP_TABLE default\n",
			    parse_name(g, a),
			    sizeof(VTK_FLOAT) == 8 ? "double" : "float");
	    else
	 	phgIOPrintf(g, "\nSCALARS %s %s %d\nLOOKUP_TABLE default\n",
			    parse_name(g, a),
			    sizeof(VTK_FLOAT) == 8 ? "double" : "float", dim);
	    buffer_pos = 0;
	    ForAllElements(g, s) {
		values = DofElementData(dof, s->index);
		for (k = 0; k < dim; k++) {
		    VTK_FLOAT v = *(values++);
		    BigEndianAppend(&v, sizeof(v), 1);
		}
	    }
	    phgIOWrite(g, buffer, sizeof(VTK_FLOAT) * dim, nleaf,
			g->nelem_global, NULL, NULL); 
	}
    }

    /* save submesh indices as cell data */
    phgIOPrintf(g, "\nSCALARS submesh_no int 1\nLOOKUP_TABLE default\n");
    buffer_pos = 0;
    ForAllElements(g, s) {
	a = g->rank;
	BigEndianAppend(&a, sizeof(a), 1);
    }
    phgIOWrite(g, buffer, sizeof(a), nleaf, g->nelem_global, NULL, NULL); 

    phgIOClose();

    phgFree(buffer);
    buffer = NULL;

    return;
}

const char *
phgExportVTKn(GRID *g, const char *fn, int ndof, DOF **dofs)
/* Outputs mesh in VTK format, returns the output filename.
 * The variable part of the arguments is a NULL terminated list of DOFs
 * to save to the VTK file */
{
    int i, j, k;
    DOF *dof;

    assert(fn != NULL && Dim == 3);
    if (g == NULL)
	return NULL;

    /* output fn */
    filename = fn;

    /* list of DOFs */
    for (i = 0, k = 0; i < ndof; i++) {
	if ((dof = dofs[i])->type == DOF_ANALYTIC) {
	    if (g->rank == 0) {
		phgWarning("DOF type 'DOF_ANALYTIC' is not supported by "
			   "phgExportVTK(). \n");
		phgWarning("DOF '%s' will not be saved.\n", dof->name);
	    }
	    continue;
	}
	if (i >= 256) {
	    if (g->rank == 0)
		phgWarning("only first 255 DOFs are saved to VTK file.");
	    break;
	}
	/* test for missing NULL terminator */
	if (!phgDofIsValid(g, dof)) {
	    if (g->rank == 0)
		phgWarning("phgExportVTK: missing NULL terminator"
			   " in argument list.\n");
	    break;
	}
	if ((j = phgDofPointer2Index(dof)) < 0)
	    break;
	if (j > 255)
	    phgWarning("susperious DOF index: %d!\n", j);
	dof_ids[++k] = j;
    }
    dof_ids[0] = k;

    phgExportVTK_(g);

    return filename;
}

const char *
phgExportVTK(GRID *g, const char *fn, DOF *dof, ...)
{
    int ndof;
    DOF **dofs = phgAlloc(256 * sizeof(*dofs));
    va_list ap;

    va_start(ap, dof);
    for (ndof = 0; ndof < 256; ndof++) {
	if (dof == NULL)
	    break;
	dofs[ndof] = dof;
	dof = va_arg(ap, DOF *);
    }
    va_end(ap);

    fn = phgExportVTKn(g, fn, ndof, dofs);
    phgFree(dofs);

    return fn;
}
