/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: quad_test.c,v 1.9 2007/01/11 14:18:37 zlb Exp $
 *
 * Numerical quadrature test. */

#include "phg.h"
#include <stdlib.h>
#include <math.h>

int
main(int argc, char *argv[])
{
    QUAD *quad;
    int i, p, np;
    FLOAT a, b;
    const FLOAT *pw, *pp;

    phgInit(&argc, &argv);

    /* 1D quadrature lambda0^p + lambda1^p */
    printf("\n========================== 1D quadrature rules.\n");
    for (p = 0; p <= 30; p++) {
	quad = phgQuadGetQuad1D(p);
	np = quad->npoints;
	pp = quad->points;
	pw = quad->weights;
	b = 0.;
	for (i = 0; i < np; i++, pp += 2)
	    b += (Pow(pp[0],p) + Pow(pp[1],p)) * *(pw++);
	/* analytic: \int lambda^p = 1 / (p + 1) */
	a = 2. / (FLOAT)(p + 1) * 1.;
	printf("    order %2d, points %4d, result = %le, rel. error = %le\n",
		p, np, (double)b, (double)Fabs((a - b) / b));
    }

    /* 2D quadrature lambda0^p + lambda1^p + lambda2^p */
    printf("\n========================== 2D quadrature rules.\n");
    for (p = 0; p <= 30; p++) {
	quad = phgQuadGetQuad2D(p);
	np = quad->npoints;
	pp = quad->points;
	pw = quad->weights;
	b = 0.;
	for (i = 0; i < np; i++, pp += 3)
	    b += (Pow(pp[0],p) + Pow(pp[1],p) + Pow(pp[2],p)) * *(pw++);
	/* analytic: \int lambda^p = 1 / ((p + 1) * (p + 2)) */
	a = 3. / (FLOAT)((p + 1) * (p + 2)) * 2.;
	printf("    order %2d, points %4d, result = %le, rel. error = %le\n",
		p, np, (double)b, (double)Fabs((a - b) / b));
    }

    /* 3D quadrature lambda0^p + lambda1^p + lambda2^p + lambda3^p */
    printf("\n========================== 3D quadrature rules.\n");
    for (p = 0; p <= 30; p++) {
	quad = phgQuadGetQuad3D(p);
	np = quad->npoints;
	pp = quad->points;
	pw = quad->weights;
	b = 0.;
	for (i = 0; i < np; i++, pp += 4)
	    b += (Pow(pp[0],p) + Pow(pp[1],p) + Pow(pp[2],p) + Pow(pp[3],p)) *
		 *(pw++);
	/* analytic: \int lambda^p = 1 / ((p + 1) * (p + 2) * (p + 3)) */
	a = 4. / (FLOAT)((p + 1) * (p + 2) * (p + 3)) * 6.;
	printf("    order %2d, points %4d, result = %le, rel. error = %le\n",
		p, np, (double)b, (double)Fabs((a - b) / b));
    }

    phgFinalize();

    return 0;
}
